# def test_storage_list(xvc_repo_with_dir):
#     assert False
#
#
# def test_storage_remove(xvc_repo_with_dir):
#     assert False
#
#
# def test_storage_new(xvc_repo_with_dir):
#     assert False
#
#
# def test_storage_new_local(xvc_repo_with_dir):
#     assert False
#
#
# def test_storage_new_generic(xvc_repo_with_dir):
#     assert False
#
#
# def test_storage_new_rsync(xvc_repo_with_dir):
#     assert False
#
#
# def test_storage_new_s3(xvc_repo_with_dir):
#     assert False
#
#
# def test_storage_new_minio(xvc_repo_with_dir):
#     assert False
#
#
# def test_storage_new_digital_ocean(xvc_repo_with_dir):
#     assert False
#
#
# def test_storage_new_r2(xvc_repo_with_dir):
#     assert False
#
#
# def test_storage_new_gcs(xvc_repo_with_dir):
#     assert False
#
#
# def test_storage_new_wasabi(xvc_repo_with_dir):
#     assert False
#
#
# def test_storage_new_help(xvc_repo_with_dir):
#     assert False
#

# def test_file_send(xvc_repo_with_dir):
#     assert False
#
#
# def test_file_bring(xvc_repo_with_dir):
#     assert False
#
#
# def test_file_share(xvc_repo_with_dir):
#     assert False
#
# def test_file_remove_from_storage(xvc_repo_with_dir):
#     assert False
#
