//! Xvc basic types used across the whole program
pub mod diff;
pub mod hashalgorithm;
pub mod recheckmethod;
pub mod textorbinary;
pub mod xvcdigest;
pub mod xvcfiletype;
pub mod xvcmetadata;
pub mod xvcpath;
pub mod xvcroot;
